; GreatCircle.asm
; Greg Lyon
; 2007

INCLUDE \masm32\include\masm32rt.inc

;---------------------------------------
; Macros used by the BearingDistance procedure
;---------------------------------------
fDegToRad MACRO
    ;; degrees in in ST(0)
    ;; radians out in ST(0)
    fld DegToRad
    fmul
ENDM    

fRadToDeg MACRO
    ;; radians in in ST(0)
    ;; degrees out in ST(0)
    fld RadToDeg
    fmul
ENDM

fAcos MACRO
    ;; x in in ST(0) 
    ;; radians out in ST(0)
    fld     st(0)           ;;ST(0) = cos(x), ST(1) = cos(x)
    fmul    st(0), st(0)    ;;ST(0) = cos2(x), ST(1) = cos(x)
    fld1                    ;;ST(0) = 1.0, ST(1) = cos2(x), ST(2) = cos(x)
    fsubr                   ;;ST(0) = 1 - cos2(x) = sin2(x), ST(1) = cos(x)
    fsqrt                   ;;ST(0) = sin(x), ST(1) = cos(x)
    fxch                    ;;ST(0) = cos(x), ST(1) = sin(x)
    fpatan                  ;;ST(0) = arccos[cos(x)] = x (in radians)
ENDM
;---------------------------------------

.CODE

start:

    call main
    inkey chr$(13,10,"Press any key to exit ... ")
    print chr$(13,10)    
    exit 0

;---------------------------------------
InputR10 PROC pR10:PTR REAL10
    LOCAL r8:REAL8
    invoke crt_scanf, SADD("%lf"), ADDR r8
    mov eax, pR10
    finit
    fld r8
    fstp REAL10 PTR [eax]
    ret
InputR10 ENDP
;---------------------------------------
BearingDistance PROC USES ebx pLat1:PTR REAL10, pLon1:PTR REAL10, pLat2:PTR REAL10, pLon2:PTR REAL10, pBearing:PTR REAL10, pDistance:PTR REAL10
    ; 
    ; Calculate the bearing (degrees) and distance (kilometers) between two latitudes and longitudes.
    ; See the included 'GreatCircle.txt' text file for a description of the formula. 
    ;
    ; Macros used by the BearingDistance procedure:
    ;
    ;fDegToRad MACRO
    ;    ;; degrees in in ST(0)
    ;    ;; radians out in ST(0)
    ;    fld DegToRad
    ;    fmul
    ;ENDM
    ;
    ;fRadToDeg MACRO
    ;    ;; radians in in ST(0)
    ;    ;; degrees out in ST(0)
    ;    fld RadToDeg
    ;    fmul
    ;ENDM
    ;
    ;fAcos MACRO
    ;    ;; x in in ST(0)
    ;    ;; radians out in ST(0)
    ;    fld     st(0)           ;;ST(0) = cos(x), ST(1) = cos(x)
    ;    fmul    st(0), st(0)    ;;ST(0) = cos2(x), ST(1) = cos(x)
    ;    fld1                    ;;ST(0) = 1.0, ST(1) = cos2(x), ST(2) = cos(x)
    ;    fsubr                   ;;ST(0) = 1 - cos2(x) = sin2(x), ST(1) = cos(x)
    ;    fsqrt                   ;;ST(0) = sin(x), ST(1) = cos(x)
    ;    fxch                    ;;ST(0) = cos(x), ST(1) = sin(x)
    ;    fpatan                  ;;ST(0) = arccos[cos(x)] = x (in radians)
    ;ENDM
    
    ALIGN 16
    LOCAL dLon_sin:REAL10
    ALIGN 16
    LOCAL d_sin:REAL10
    ALIGN 16
    LOCAL d_cos:REAL10
    ALIGN 16
    LOCAL Lat1_sin:REAL10
    ALIGN 16
    LOCAL Lat1_cos:REAL10
    ALIGN 16
    LOCAL Lat2_sin:REAL10

    .CONST
    
      ALIGN 16
      DegToRad      REAL10    0.017453292519943295769236907684
      ALIGN 16
      RadToDeg      REAL10   57.295779513082320876798154814105
      ALIGN 16
      KilometerMult REAL10 6378.7
      ALIGN 16
      ext360        REAL10  360.0

    .CODE        
    
      ; *** Distance ***

      mov eax, pLat1
      mov ebx, pLat2
      mov ecx, pLon1
      mov edx, pLon2

      finit
      fld REAL10 PTR [eax]      ; load Lat1
      fDegToRad                 ; convert degrees to radians
      fld REAL10 PTR [ebx]      ; load Lat2
      fDegToRad                 ; convert degrees to radians
      fld REAL10 ptr [ecx]      ; load Lon1
      fld REAL10 PTR [edx]      ; load Lon2
                                ; ST(0) = Lon2 as degrees
                                ; ST(1) = Lon1 as degrees
                                ; ST(2) = Lat2 as radians
                                ; ST(3) = Lat1 as radians
      fsub
      fDegToRad                 ; ST(0) = dLon = Lon1 - Lon2 as radians
                                ; ST(1) = Lat2 as radians
                                ; ST(2) = Lat1 as radians
      fsincos                   ; ST(0) = cos(dLon), ST(1) = sin(dLon)
      fxch                      ; ST(0) = sin(dLon), ST(1) = cos(dLon)
      fstp dLon_sin             ; save sin(dLon) for Bearing
                                ; ST(0) = cos(dLon)
                                ; ST(1) = Lat2 as radians
                                ; ST(2) = Lat1 as radians
      fxch                      ; ST(0) = Lat2 as radians
                                ; ST(1) = cos(dLon)
                                ; ST(2) = Lat1 as radians
      fsincos                   ; ST(0) = cos(Lat2), ST(1) = sin(Lat2)
      fxch                      ; ST(0) = sin(Lat2), ST(1) = cos(Lat2)
      fst st(4)                 ; save sin(Lat2) to st(4)
      fstp Lat2_sin             ; save sin(Lat2) for Bearing
                                ; ST(0) = cos(Lat2)
                                ; ST(1) = cos(dLon)
                                ; ST(2) = Lat1 as radians
                                ; ST(3) = Lat2_sin
      fxch st(2)                ; ST(0) = Lat1 as radians
                                ; ST(1) = cos(dLon)
                                ; ST(2) = cos(Lat2)
                                ; ST(3) = Lat2_sin
      fsincos                   ; ST(0) = cos(Lat1), ST(1) = sin(Lat1) 
      fld st(0)                 ; duplicate cos(Lat1)
      fstp Lat1_cos             ; save cos(Lat1) for Bearing
      fxch                      ; ST(0) = sin(Lat1), ST(1) = cos(Lat1)
      fst st(5)                 ; save sin(Lat1) to ST(5)
      fstp Lat1_sin             ; save sin(Lat1) for Bearing
                                ; ST(0) = cos(Lat1)
                                ; ST(1) = cos(dLon)
                                ; ST(2) = cos(Lat2)
                                ; ST(3) = Lat2_sin
                                ; ST(4) = Lat1_sin
      fmul
      fmul                      ; ST(0) = cos(Lat1) * cos(Lat2) * cos(dLon)
                                ; ST(1) = Lat2_sin
                                ; ST(2) = Lat1_sin
      fxch st(2)                ; ST(0) = Lat1_sin
                                ; ST(1) = Lat2_sin
                                ; ST(2) = cos(Lat1) * cos(Lat2) * cos(dLon)
      fmul                      ; ST(0) = sin(Lat1) * sin(Lat2)
                                ; ST(1) = cos(Lat1) * cos(Lat2) * cos(dLon)
      fadd                      ; ST(0) = (sin(Lat1) * sin(Lat2)) + (cos(Lat1) * cos(Lat2) * cos(dLon))
      fAcos                     ; ST(0) = d = acos( (sin(Lat1) * sin(Lat2)) + (cos(Lat1) * cos(Lat2) * cos(dLon)) )
      fld st(0)                 ; duplicate d
      fsincos                    ; ST(0) = cos(d), ST(1) = sin(d)
      fstp d_cos                ; save cos(d) for Bearing
      fstp d_sin                ; save sin(d) for Bearing
      fld KilometerMult         ; load the multiplier for kilometers
      fmul
      mov eax, pDistance        ; get the pointer to the distance
      fstp REAL10 PTR [eax]     ; save the distance

      ; *** Bearing ***

      fld Lat2_sin              ; load sin(Lat2)
      fld Lat1_sin              ; load sin(Lat1)
      fld d_cos                 ; load cos(d)
      fld Lat1_cos              ; load cos(Lat1)
      fld d_sin                 ; load sin(d)
                                ; ST(0) = sin(d)
                                ; ST(1) = cos(Lat1)
                                ; ST(2) = cos(d)
                                ; ST(3) = sin(Lat1)
                                ; ST(4) = sin(Lat2)
      fmul                      ; ST(0) = cos(Lat1) * sin(d)
                                ; ST(1) = cos(d)
                                ; ST(2) = sin(Lat1)
                                ; ST(3) = sin(Lat2)
      fxch st(2)                ; ST(0) = sin(Lat1)
                                ; ST(1) = cos(d)
                                ; ST(2) = cos(Lat1) * sin(d)
                                ; ST(3) = sin(Lat2)
      fmul                      ; ST(0) = sin(Lat1) * cos(d)
                                ; ST(1) = cos(Lat1) * sin(d)
                                ; ST(2) = sin(Lat2)
      fxch                      ; ST(0) = cos(Lat1) * sin(d)
                                ; ST(1) = sin(Lat1) * cos(d)
                                ; ST(2) = sin(Lat2)
      fxch st(2)                ; ST(0) = sin(Lat2)
                                ; ST(1) = sin(Lat1) * cos(d)
                                ; ST(2) = cos(Lat1) * sin(d)
      fsubr                     ; ST(0) = sin(Lat2) - (sin(Lat1) * cos(d))
                                ; ST(1) = cos(Lat1) * sin(d)
      fdivr                     ; ST(0) = (sin(Lat2) - (sin(Lat1) * cos(d))) / (cos(Lat1) * sin(d))
      fAcos                     ; ST(0) = b = acos ( (sin(Lat2) - (sin(Lat1) * cos(d))) / (cos(Lat1) * sin(d)) )
      fRadToDeg                 ; convert radians to degrees
      fld dLon_sin              ; load sin(dLon)
                                ; is it positive?
      ftst                      ; compare the value of ST(0) to +0.0
      fstsw ax                  ; copy the Status Word containing the result to AX
      fwait                     ; insure the previous instruction is completed
      sahf                      ; transfer the condition codes to the CPU's flag register
      ;jpe error_handler        ; the comparison was indeterminate
                                ; this condition should be verified first
                                ; then only two of the next three conditional jumps
                                ; should become necessary, in whatever order is preferred,
                                ; the third jump being replaced by code to handle that case
      ja st0_positive           ; when all flags are 0
      ;jb st0_negative          ; only the C0 bit (CF flag) would be set if no error
      ;jz st0_zero              ; only the C3 bit (ZF flag) would be set if no error
      jmp @F

    st0_positive:               ; if sin(dLon) is positive
      fld ext360                ; load 360.0
      fsubr st(2), st(0)        ; subtract bearing from 360.0
      fstp st(0)                ; pop st(0)

    @@:
      fstp st(0)                ; pop st(0)
      mov eax, pBearing         ; get the pointer to the bearing
      fstp REAL10 PTR [eax]     ; save the bearing

      ret
    
BearingDistance ENDP
;---------------------------------------
main PROC

    ALIGN 16
    LOCAL extLat1:REAL10
    ALIGN 16    
    LOCAL extLon1:REAL10
    
    ALIGN 16    
    LOCAL extLat2:REAL10
    ALIGN 16    
    LOCAL extLon2:REAL10

    ALIGN 16
    LOCAL extKilometers:REAL10
    ALIGN 16
    LOCAL extStatuteMiles:REAL10
    Align 16
    LOCAL extNauticalMiles:REAL10
    ALIGN 16
    LOCAL extBearing:REAL10

    .CONST

      ALIGN 16
      KmToStatuteMile  REAL10 0.62137119223733396961743418436332
      ALIGN 16
      KmToNauticalMile REAL10 0.53995680345572354211663066954644

    .CODE

      print chr$("Great Circle Bearing and Distance",13,10,13,10)
      print chr$("When entering latitude and longitude,",13,10)
      print chr$("north and east are positive,",13,10)
      print chr$("south and west are negative.",13,10)
      print chr$("Use decimal degrees (ddd.ddd).",13,10)
      print chr$(13,10)
      
      print chr$("Enter latitude 1:  ")
      INVOKE InputR10, ADDR extLat1
      print chr$("Enter longitude 1: ")
      INVOKE InputR10, ADDR extLon1
      print chr$(13,10)
      
      print chr$("Enter latitude 2:  ")
      INVOKE InputR10, ADDR extLat2
      print chr$("Enter longitude 2: ")
      INVOKE InputR10, ADDR extLon2
      
      INVOKE BearingDistance, ADDR extLat1, ADDR extLon1, ADDR extLat2, ADDR extLon2, ADDR extBearing, ADDR extKilometers

      ; convert kilometers to miles
      finit
      fld extKilometers
      fld st(0)
      fld KmToStatuteMile
      fmul
      fstp extStatuteMiles
      fld KmToNauticalMile
      fmul
      fstp extNauticalMiles

      print chr$(13,10)
      print chr$("Bearing:  ")
      print real10$(extBearing)
      print chr$(" degrees",13,10)

      print chr$("Distance: ")
      print real10$(extKilometers)
      print chr$(" kilometers",13,10)

      print chr$("          ")
      print real10$(extStatuteMiles)
      print chr$(" statute miles",13,10)

      print chr$("          ")
      print real10$(extNauticalMiles)
      print chr$(" nautical miles",13,10)


      xor eax, eax
      ret
main ENDP
;---------------------------------------
END start 
